from socket import *
import time
import sys
METRIC="EUCLIDEAN"


class Movie:
	def __init__(self,line):
		line = line.split("\t")
		self.ratings={}
		for x in line[1:]:
			
			x=x.split(",")
			if len(x)!=2:
				print x
			else:
				self.ratings[int(x[0])]=float(x[1])
		self.id=int(line[0])
	
	def findClosest(self, clusters):
		error=()
		nearest=None
		for cluster in clusters:
			if cluster.ratings==self.ratings:
				print "same"
			d=self.distance(cluster)
			if d==0.0:
				print "0 Distance"
			if d<error:
				nearest=cluster
				error=d
		nearest.contains.append(self)
		nearest.error+=error
		
	def distance(self,cluster):
		if METRIC=="EUCLIDEAN":
			return self.eucDistance(cluster)
		else:
			return self.cosDistance(cluster)
			
	def eucDistance(self,cluster):
		sum=0.0
		for id in cluster.ratings:
			if not id in self.ratings:
				sum+=(cluster.ratings[id])**2
		for id in self.ratings:
			if id in cluster.ratings:
				
				sum+=(self.ratings[id]-cluster.ratings[id])**2
			else:
				sum+=(self.ratings[id])**2
		return sum
		
		
class Cluster:
	def __init__(self,line):
		line=line.split("\t")
		self.ratings={}
		for x in line:
			x=x.split(",")
			if len(x)!=2:
				print x
			else:
				self.ratings[int(x[0])]=float(x[1])
		self.error=0.0
		self.contains=[]
		
	def distance(self,movie):
		sum=0
		if METRIC=="EUCLIDEAN":
			for id in movie.ratings.keys():
				if not id in self.ratings:
					sum+=movie.ratings[id]**2
			for id in self.ratings.keys():
				if not id in movie.ratings:
					sum+=self.ratings[id]**2
				else:
					sum+=(self.ratings[id]-movie.ratings[id])**2
		if METRIC=="COSINE":
			for id in movie.ratings.keys():
				if id in self.ratings:
					sum-=movie.ratings[id]*self.ratings[id]
		return sum
		
	def update(self):
		self.ratings={}
		for movie in self.contains:
			for id in movie.ratings.keys():
				if id in self.ratings:
					self.ratings[id]+=movie.ratings[id]
				else:
					self.ratings[id]=movie.ratings[id]
		for id in self.ratings.keys():
			self.ratings[id]/=len(self.contains)
			
	def __str__(self):
		rv=str(self.error)+"\t"+str(len(self.contains))
		for id in self.ratings:
			rv+="\t"+str(id)+","+str(self.ratings[id])
		return rv
class kmeansClient:
	def __init__(self,serverHost,serverPort):
		self.soc= socket(AF_INET, SOCK_STREAM)
		self.soc.connect((serverHost, serverPort))

		
	def loadPoints(self):
		print "LOADING"
		self.movies=[]
		data="GO"
		while data!="STOP":
			self.soc.send("READY\n")
			if data!="GO":
				self.movies.append(Movie(data))
			data=self.readLine()
		print "LOADED"
			
	
	def readLine(self):
		data=self.soc.recv(1024)
		while data[-1]!="\n":
			data+=self.soc.recv(1024)
		return data.strip()
			
	def loadClusters(self):
		print "LOADING CLUSTERS"
		self.clusters=[]
		data="GO"
		while data!="STOP":
			self.soc.send("READY\n")
			if data!="GO":
				self.clusters.append(Cluster(data))
			data=self.readLine()
			
			
	def kmeansStep(self):
		print "KMEANS"
		error=0
		for movie in self.movies:
			movie.findClosest(self.clusters)
			
	def sendClusters(self):
		print "SENDING CLUSTERS"
		data=""
		for c in self.clusters:
			self.readLine()
			self.soc.send(str(c)+"\n")
			
			
	def start(self):
		self.loadPoints()
		data="GO"
		while data!="STOP":
			self.loadClusters()
			self.kmeansStep()
			self.sendClusters()
			self.readLine()
		print "DONE"
			
			
port = int(sys.argv[1])
client=kmeansClient("localhost",port)
client.start()

for c in client.clusters:
	print c.error,len(c.contains)
	time.sleep(1)

	