/* The framework for manipulating the netflix dataset */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <time.h>

#include "netflixcore.h"

/* Some utility functions */

/* random number from 0.0 to 1.0 */
double rand_0_1() {
  return (double)rand() / ((double)(RAND_MAX)+(double)(1));
}

static int cmpratings(const void* a, const void* b) {
  rating *r1 = (rating*)a;
  rating *r2 = (rating*)b;
  if(r1->id < r2->id)
    return -1;
  if(r1->id > r2->id)
    return 1;
  return 0;
}

int datetoint(int year, int month, int day) {
  /* DO THIS */
  return 0;
}

/* ALSO, DO THE REVERSE */

/* Implementations for functions in netflixcore.h */
int nf_allocate() {
  /* File variables */
  FILE *rpm, *mappings;
  const char *rpmname = "./rpm.txt";
  const char *mappingsname = "./mappings_final.txt";
  char line[100];
  
  /* Utility variables */
  int i, j, ret, itmp;
  
  /* Allocate the movie->user->rating table */
  rpm = fopen(rpmname, "r");
  if(rpm == NULL) { perror("Can't open rpm.txt"); return 1; }
  nf_ratings.mur = (rlist*)malloc(sizeof(rlist)*NMOVIES);
  if(nf_ratings.mur == NULL) { perror("Can't malloc mur"); return 1; }
  for(i = 0; i < NMOVIES; i++) {
    ret = fscanf(rpm, "%s", line);
    assert(ret != EOF);
    nf_ratings.mur[i].length = atoi(line);
    nf_ratings.mur[i].ratings = (rating*)malloc(sizeof(rating)*nf_ratings.mur[i].length);
    if(nf_ratings.mur[i].ratings == NULL) { perror("Can't malloc a row of mur"); return 1; }
    for(j = 0; j < nf_ratings.mur[i].length; j++)
      nf_ratings.mur[i].ratings[j].id = NOID;
  }
#ifndef SMALL
  /* Make sure the file is all empty now */
  while((j = fgetc(rpm)) != EOF)
    assert(isspace(j));
#endif
  fclose(rpm);

  /* Allocate the user->movie->rating table */
  mappings = fopen(mappingsname, "r");
  if(mappings == NULL) { perror("Can't open mappings.txt"); return 1; }
  nf_ratings.umr = (rlist*)malloc(sizeof(rlist)*NUSERS);
  if(nf_ratings.umr == NULL) { perror("Can't malloc umr"); return 1; }
  fscanf(mappings, "%s", line);
  fscanf(mappings, "%s", line);
  fscanf(mappings, "%s", line);
  for(i = 0; i < NUSERS; i++) {
    ret = fscanf(mappings, "%d", &itmp);
    assert(ret != EOF);
    ret = fscanf(mappings, "%d", &itmp);
    assert(ret != EOF);
    ret = fscanf(mappings, "%d", &itmp);
    assert(ret != EOF);
    nf_ratings.umr[i].length = itmp;
    nf_ratings.umr[i].ratings = (rating*)malloc(sizeof(rating)*(itmp));
    if(nf_ratings.umr[i].ratings == NULL) { perror("Can't malloc a row of umr"); return 1; }
    for(j = 0; j < nf_ratings.umr[i].length; j++)
      nf_ratings.umr[i].ratings[j].id = NOID;
  }
#ifndef SMALL
  while((j = fgetc(mappings)) != EOF)
    assert(isspace(j));
#endif
  fclose(mappings);
  
  return 0;
}

int nf_load() {
  /* File stuff */
  FILE *mfile;
  DIR *trainingset;
  struct dirent *d_ent;
  const char *trainingsetname = "/NetflixData/download/training_set_update";
  char line[100];
  
  /* Utility variables */
  int i, j, k, itmp, ret;
  int id, stars, year, month, day;
  int mtotal, utotal, realtotal;  

  mtotal = utotal = realtotal = 0;

  /* Run over the training files, filling the data structures */
  trainingset = opendir(trainingsetname);
  if(trainingset == NULL) { perror("Can't open training set"); return 1; }
  for(i = 0; i < NMOVIES; i++) {
    d_ent = readdir(trainingset);
    assert(d_ent != NULL);
    if(strcmp(".", d_ent->d_name) == 0 || strcmp("..", d_ent->d_name) == 0) {
      i--;
      continue;
    }
    
    /* A wee bit dangerous? */
    strcpy(line, trainingsetname);
    strcat(line, "/");
    strcat(line, d_ent->d_name);
    
    mfile = fopen(line, "r");
    if(mfile == NULL) { perror(line); return 1; }
    ret = fscanf(mfile, "%s", line); /* Get the header */
    assert(ret != EOF);
    line[strlen(line)-1] = 0;        /* Kill the trailing colon */
    itmp = atoi(line) - 1;           /* Movie ID - 1 */
    assert(itmp == i);               /* Dangerous: This relies on the order the directory is traversed (but it warns us anyways if it fails) */
    for(j = 0; j < nf_ratings.mur[itmp].length; j++) {
      ret = fscanf(mfile, "%s", line);
      assert(ret != EOF);
      sscanf(line, "%d,%d,%d-%d-%d", &id, &stars, &year, &month, &day);
      
      if(id < NUSERS) {
        realtotal++;

        /* Put it in mur */
        nf_ratings.mur[itmp].ratings[j].id = id;
        nf_ratings.mur[itmp].ratings[j].stars = stars;
#ifdef TEMPORAL
        nf_ratings.mur[itmp].ratings[j].date = datetoint(year, month, day);
#endif
        nf_ratings.mur[itmp].ratings[j].visible = 1;

        /* Put it in umr */
        for(k = 0; nf_ratings.umr[id].ratings[k].id != NOID; k++);
        assert(k < nf_ratings.umr[id].length);
        nf_ratings.umr[id].ratings[k].id = itmp;
        nf_ratings.umr[id].ratings[k].stars = stars;
#ifdef TEMPORAL
        nf_ratings.umr[id].ratings[k].date = datetoint(year, month, day);
#endif
        nf_ratings.umr[id].ratings[k].visible = 1;
      }
    }
    while((j = fgetc(mfile)) != EOF)
      assert(isspace(j));
    fclose(mfile);
    qsort(nf_ratings.mur[itmp].ratings, nf_ratings.mur[itmp].length, sizeof(rating), cmpratings);
    for(j = 0; j < nf_ratings.mur[itmp].length; j++) {
      if(nf_ratings.mur[itmp].ratings[j].id == NOID) {
        /* printf("shortening movie %d...\n", itmp); */
        break;
      }
    }
    nf_ratings.mur[itmp].length = j;
    mtotal += j;
  }
  if(closedir(trainingset) == -1) { perror("Can't close training set"); return 1; }
  for(i = 0; i < NUSERS; i++) {
    qsort(nf_ratings.umr[i].ratings, nf_ratings.umr[i].length, sizeof(rating), cmpratings);
    for(j = 0; j < nf_ratings.umr[i].length; j++) {
      if(nf_ratings.umr[i].ratings[j].id == NOID) {
        /* printf("shortening user %d...\n", i); */
        break;
      }
    }
    nf_ratings.umr[i].length = j;
    utotal += j;
  }

  assert(realtotal == utotal && realtotal == mtotal);

  return 0;
}

int nf_cleanup() {
  int i;
  for(i = 0; i < NMOVIES; i++) {
    free(nf_ratings.mur[i].ratings);
    nf_ratings.mur[i].ratings = NULL;
  }
  free(nf_ratings.mur);
  nf_ratings.mur = NULL;
  
  for(i = 0; i < NUSERS; i++) {
    free(nf_ratings.umr[i].ratings);
    nf_ratings.umr[i].ratings = NULL;
  }
  free(nf_ratings.umr);
  nf_ratings.umr = NULL;
  
  return 0;
}

int nf_longtest(double cutoff) {
  int i, j;
  rating *rtmp, *rtmp2;
  
  for(i = 0; i < NMOVIES; i++)
    for(j = 0; j < NUSERS; j++) {
      if(rand_0_1() < cutoff) {
        /* Look for a rating in mur */
        rtmp = nf_findrating_mu(i, j);
      
        /* Look for a rating in umr */
        rtmp2 = nf_findrating_um(j, i);
      
        /* Do they agree? */
        if(rtmp == NULL) {
          assert(rtmp2 == NULL);
        }
        else {
          assert(rtmp->id == j);
          assert(rtmp2->id == i);
#ifdef TEMPORAL
          assert(rtmp->date == rtmp2->date);
#endif
          assert(rtmp->stars == rtmp2->stars);
          assert(rtmp->visible == rtmp2->visible);
        }
      }
    }

  return 0;
}

inline rating* nf_findrating(int movie_id, int user_id) {
  rating *rtmp, key;
  /* look for it in mur */
  key.id = user_id;
  rtmp = bsearch(&key, nf_ratings.mur[movie_id].ratings, nf_ratings.mur[movie_id].length, sizeof(rating), cmpratings);
  if(rtmp == NULL || rtmp->visible == 0) {
    return NULL;
  }
  return rtmp;
}

inline rating* nf_findrating_mu(int movie_id, int user_id) {
  rating *rtmp, key;
  /* look for it in mur */
  key.id = user_id;
  rtmp = bsearch(&key, nf_ratings.mur[movie_id].ratings, nf_ratings.mur[movie_id].length, sizeof(rating), cmpratings);
  return rtmp;
}

inline rating* nf_findrating_um(int user_id, int movie_id) {
  rating *rtmp, key;
  /* look for it in mur */
  key.id = movie_id;
  rtmp = bsearch(&key, nf_ratings.umr[user_id].ratings, nf_ratings.umr[user_id].length, sizeof(rating), cmpratings);
  return rtmp;
}

inline rlist* nf_ratingsformovie(int movie_id) {
  return &nf_ratings.mur[movie_id];
}

inline rlist* nf_ratingsforuser(int user_id) {
  return &nf_ratings.umr[user_id];
}
