/* Declarations for the framework */

/* Constants */

/* Reserved ID meaning "not yet assigned" */
#define NOID 1000000

#ifdef SMALL
/* Load a reduced set of the data */
#define NMOVIES 800
#define NUSERS  19890
#else
/* Load the full set of data */
#define NMOVIES 17770
#define NUSERS  480189
#endif

/* Data structures */

/* Holds 1 rating */
typedef struct {
  unsigned int id : 20;    /* Either the movie or user id */
  unsigned int stars : 3;  /* The # of stars */
#ifdef TEMPORAL
  unsigned int date : 12;  /* Date (days after Dec 31st 1999) */
#endif
  unsigned int visible: 1; /* Flag for training/testing */
} rating;

/* Holds all the ratings for a user/movie */
typedef struct {
  int length;       /* number of ratings */
  rating* ratings;  /* array of ratings */
  void* additional; /* put your extra data structures here */
} rlist;

/* Holds all the ratings */
struct {
  rlist *mur; /* All the ratings indexed by movie, then by user */
  rlist *umr; /* All the ratings indexed by user, then by movie */
} nf_ratings;

/* Functions for manipulating the data */

/* Initializes the data structures */
int nf_allocate();

/* Loads the data from file (Takes ~5 min) */
int nf_load();

/* Clean up the data structures */
int nf_cleanup();

/* Boring test to verify loading of data */
int nf_longtest(double cutoff);

/* Retrieve a rating. Returns NULL if rating doesn't exist or is invisible. */
inline rating* nf_findrating(int movie_id, int user_id);

/* Retrieve a rating regardless of visibility. You specify the order. */
inline rating* nf_findrating_mu(int movie_id, int user_id); 
inline rating* nf_findrating_um(int user_id, int movie_id);

/* Retrieve an rlist */
inline rlist* nf_ratingsformovie(int movie_id);
inline rlist* nf_ratingsforuser(int user_id);

/* Train, probe, test, etc. */
/* DO THESE */

/* Utility stuff that is useful */

/* Random number from 0.0 to 1.0 */
double rand_0_1();
